/* QWalk (qwstatistics.c) 
 * Copyright (C) 2007  Franklin Marquezino
 *                                                                                                                           
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *                                                                                                                           
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA
 */

#include<stdio.h>
#include<stdlib.h>
#include<complex.h>
#include<math.h>
#include "qwstatistics.h"
#include "qwconsts.h"
#include "qwoptions_io.h"


statistics_t getStatisticsFromState1D(double complex **matrix, 
				      options1D_t opts, int iteration){
  statistics_t stat;
  double fstMoment, secMoment;
  int m;
  
  if(opts.max<1){
    stat.iteration=-1;
    return stat;
  }
  if(!matrix){
    stat.iteration=-2;
    return stat;
  }
  if(iteration<0){
    stat.iteration=-3;
    return stat;
  }

  fstMoment = 0.0;
  secMoment = 0.0;

  if(opts.lattType == LINE_LATT){
    for(m=-opts.max; m<=opts.max; m++){
      double prob;
      int j;

      prob = 0.0;
      for(j=0; j<2; j++)
	prob += matrix[j][opts.max+m]*conj(matrix[j][opts.max+m]);

      fstMoment += m*prob;
      secMoment += m*m*prob;
    }
  }
  else{
    for(m=0; m<opts.max; m++){
      double prob;
      int j;

      prob = 0.0;
      for(j=0; j<2; j++)
	prob += matrix[j][m]*conj(matrix[j][m]);

      fstMoment += m*prob;
      secMoment += m*m*prob;
    }
  }

  stat.iteration = iteration;
  stat.meanX = fstMoment;
  stat.meanY = 0.0;
  stat.variance = secMoment - (fstMoment)*(fstMoment);

  return stat;
}


statistics_t getStatisticsFromState2D(double complex ****matrix, int max, 
				      int iteration){
  statistics_t stat;
  double fstMomentX, secMomentX, varianceX;
  double fstMomentY, secMomentY, varianceY;
  int m,n;
  
  if(max<1){
    stat.iteration=-1;
    return stat;
  }
  if(!matrix){
    stat.iteration=-2;
    return stat;
  }
  if(iteration<0){
    stat.iteration=-3;
    return stat;
  }

  fstMomentX = 0.0;
  secMomentX = 0.0;
  fstMomentY = 0.0;
  secMomentY = 0.0;

  for(m=-max; m<=max; m++){
    for(n=-max; n<=max; n++){

      double prob;
      int j,k;

      prob = 0.0;
      for(j=0; j<2; j++){
	for(k=0; k<2; k++){
	  prob += matrix[j][k][max+m][max+n]*conj(matrix[j][k][max+m][max+n]);
	}
      }

      fstMomentX += m*prob;
      secMomentX += m*m*prob;
      fstMomentY += n*prob;
      secMomentY += n*n*prob;     
    }
  }
  
  varianceX = secMomentX - fstMomentX*fstMomentX;
  varianceY = secMomentY - fstMomentY*fstMomentY;  

  stat.iteration = iteration;
  stat.meanX = fstMomentX;
  stat.meanY = fstMomentY;

  /* To calculate the variance in the 2D simulation we calculate the 
   * variances of both X and Y position and add them. It simplifies
   * the calculation and gives a good approximation of the expected
   * behaviour.
   */
  stat.variance = varianceX + varianceY;

  return stat;
}



statistics_t getStatisticsFromProb1D(double *matrix, int max, 
				     int iteration){
  statistics_t stat;
  double fstMoment, secMoment;
  int m;
  
  if(max<1){
    stat.iteration=-1;
    return stat;
  }
  if(!matrix){
    stat.iteration=-2;
    return stat;
  }
  if(iteration<0){
    stat.iteration=-3;
    return stat;
  }

  fstMoment = 0.0;
  secMoment = 0.0;

  for(m=-max; m<=max; m++){
    fstMoment += m*matrix[max+m];
    secMoment += m*m*matrix[max+m];
  }

  stat.iteration = iteration;
  stat.meanX = fstMoment;
  stat.meanY = 0.0;
  stat.variance = secMoment - (fstMoment)*(fstMoment);

  return stat;
}



statistics_t getStatisticsFromProb2D(double **matrix, int max, int iteration){
  statistics_t stat;
  double fstMomentX, secMomentX, varianceX;
  double fstMomentY, secMomentY, varianceY;
  int m,n;
  
  if(max<1){
    stat.iteration=-1;
    return stat;
  }
  if(!matrix){
    stat.iteration=-2;
    return stat;
  }
  if(iteration<0){
    stat.iteration=iteration;
    return stat;
  }


  fstMomentX = 0.0;
  secMomentX = 0.0;
  fstMomentY = 0.0;
  secMomentY = 0.0;

  for(m=-max; m<=max; m++){
    for(n=-max; n<=max; n++){
      fstMomentX += m*matrix[max+m][max+n];
      secMomentX += m*m*matrix[max+m][max+n];
      fstMomentY += n*matrix[max+m][max+n];
      secMomentY += n*n*matrix[max+m][max+n];     
    }
  }
  varianceX = secMomentX - fstMomentX*fstMomentX;
  varianceY = secMomentY - fstMomentY*fstMomentY;  

  stat.iteration = iteration;
  stat.meanX = fstMomentX;
  stat.meanY = fstMomentY;
  stat.variance = varianceX + varianceY;

  return stat;
}
